@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Olzhas Rakhimov
+ * Copyright (C) 2014-2017 Olzhas Rakhimov
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,12 +27,17 @@
 #include <utility>
 #include <vector>
 
+#include <boost/multi_index_container.hpp>
+#include <boost/multi_index/global_fun.hpp>
+#include <boost/multi_index/hashed_index.hpp>
 #include <boost/noncopyable.hpp>
+#include <boost/variant.hpp>
 #include <libxml++/libxml++.h>
 
 #include "ccf_group.h"
 #include "element.h"
 #include "event.h"
+#include "event_tree.h"
 #include "expression.h"
 #include "expression/constant.h"
 #include "fault_tree.h"
@@ -69,14 +74,22 @@ class Initializer : private boost::noncopyable {
 
  private:
   /// Convenience alias for expression extractor function types.
-  using ExtractorFunction = std::function<ExpressionPtr(const xmlpp::NodeSet&,
-                                                        const std::string&,
-                                                        Initializer*)>;
+  using ExtractorFunction = std::unique_ptr<Expression> (*)(
+      const xmlpp::NodeSet&, const std::string&, Initializer*);
   /// Map of expression names and their extractor functions.
   using ExtractorMap = std::unordered_map<std::string, ExtractorFunction>;
-
-  /// Expressions mapped to their extraction functions.
-  static const ExtractorMap kExpressionExtractors_;
+  /// Container for late defined constructs.
+  template <class... Ts>
+  using TbdContainer =
+      std::vector<std::pair<boost::variant<Ts*...>, const xmlpp::Element*>>;
+  /// Container with full paths to elements.
+  ///
+  /// @tparam T  The element type.
+  template <typename T>
+  using PathTable = boost::multi_index_container<
+      T*, boost::multi_index::indexed_by<
+              boost::multi_index::hashed_unique<boost::multi_index::global_fun<
+                  const T*, std::string, &GetFullPath>>>>;
 
   /// @tparam T  Type of an expression.
   /// @tparam N  The number of arguments for the expression.
@@ -86,6 +99,23 @@ class Initializer : private boost::noncopyable {
   template <class T, int N>
   struct Extractor;
 
+  /// Expressions mapped to their extraction functions.
+  static const ExtractorMap kExpressionExtractors_;
+
+  /// Calls Extractor with an appropriate N to construct the expression.
+  ///
+  /// @tparam T  Type of an expression.
+  ///
+  /// @param[in] args  A vector of XML elements containing the arguments.
+  /// @param[in] base_path  Series of ancestor containers in the path with dots.
+  /// @param[in,out] init  The host Initializer.
+  ///
+  /// @returns The new extracted expression.
+  template <class T>
+  static std::unique_ptr<Expression> Extract(const xmlpp::NodeSet& args,
+                                             const std::string& base_path,
+                                             Initializer* init);
+
   /// Checks if all input files exist on the system.
   ///
   /// @param[in] xml_files  The XML input files.
@@ -133,15 +163,51 @@ class Initializer : private boost::noncopyable {
   /// @throws ValidationError  The elements contain undefined dependencies.
   void ProcessTbdElements();
 
-  /// Attaches attributes and a label to the elements of the analysis.
-  /// These attributes are not XML attributes
-  /// but the Open-PSA format defined arbitrary attributes
-  /// and a label that can be attached to many analysis elements.
+  /// Registers an element into the model.
   ///
-  /// @param[in] element_node  XML element.
-  /// @param[out] element  The object that needs attributes and label.
-  void AttachLabelAndAttributes(const xmlpp::Element* element_node,
-                                Element* element);
+  /// @tparam T  A pointer type to the element.
+  ///
+  /// @param[in] element  The initialized element ready to be added into models.
+  /// @param[in] xml_element  The related XML element for error messages.
+  ///
+  /// @throws ValidationError  Issues with adding the element into the model.
+  template <class T>
+  void Register(T&& element, const xmlpp::Element* xml_element);
+
+  /// Constructs and registers an element in the model.
+  ///
+  /// @tparam T  The Element type to be specialized for.
+  ///
+  /// @param[in] xml_node  XML node defining the element.
+  /// @param[in] base_path  Series of ancestor containers in the path with dots.
+  /// @param[in] base_role  The parent container's role.
+  ///
+  /// @returns Pointer to the registered element.
+  ///
+  /// @throws ValidationError  Issues with the new element or registration.
+  template <class T>
+  T* Register(const xmlpp::Element* xml_node, const std::string& base_path,
+              RoleSpecifier base_role);
+
+  /// Adds additional data to element definition
+  /// after processing all the input files.
+  ///
+  /// @tparam T  The Element type to be specialized for.
+  ///
+  /// @param[in] xml_node  XML element defining the element.
+  /// @param[in,out] element  Registered element ready to be defined.
+  ///
+  /// @throws ValidationError  Issues with the additional data.
+  template <class T>
+  void Define(const xmlpp::Element* xml_node, T* element);
+
+  /// Defines an event tree for the analysis.
+  ///
+  /// @param[in] et_node  XML element defining the event tree.
+  ///
+  /// @throws ValidationError  There are issues with registering and defining
+  ///                          the event tree and its data.
+  void DefineEventTree(const xmlpp::Element* et_node);
 
   /// Defines a fault tree for the analysis.
   ///
@@ -186,25 +252,6 @@ class Initializer : private boost::noncopyable {
   /// @param[in] model_data  XML node with model data description.
   void ProcessModelData(const xmlpp::Element* model_data);
 
-  /// Registers a gate for later definition.
-  ///
-  /// @param[in] gate_node  XML element defining the gate.
-  /// @param[in] base_path  Series of ancestor containers in the path with dots.
-  /// @param[in] container_role  The parent container's role.
-  ///
-  /// @returns Pointer to the registered gate.
-  ///
-  /// @throws ValidationError  An event with the same name is already defined.
-  GatePtr RegisterGate(const xmlpp::Element* gate_node,
-                       const std::string& base_path,
-                       RoleSpecifier container_role);
-
-  /// Defines a gate for this analysis.
-  ///
-  /// @param[in] gate_node  XML element defining the gate.
-  /// @param[in,out] gate  Registered gate ready to be defined.
-  void DefineGate(const xmlpp::Element* gate_node, Gate* gate);
-
   /// Creates a Boolean formula from the XML elements
   /// describing the formula with events and other nested formulas.
   ///
@@ -217,118 +264,53 @@ class Initializer : private boost::noncopyable {
   FormulaPtr GetFormula(const xmlpp::Element* formula_node,
                         const std::string& base_path);
 
-  /// Processes the arguments of a formula with nodes and formulas.
+  /// Processes event tree branch instructions and target from XML data.
   ///
-  /// @param[in] formula_node  The XML element with children as arguments.
-  /// @param[in] base_path  Series of ancestor containers in the path with dots.
-  /// @param[in,out] formula  The formula to be defined by the arguments.
+  /// @param[in] xml_nodes  The XML element data of the branch.
+  /// @param[in,out] event_tree  The host event tree as a context container.
+  /// @param[out] branch  The branch to be defined with instructions and target.
   ///
-  /// @throws ValidationError  Repeated arguments are identified.
-  void ProcessFormula(const xmlpp::Element* formula_node,
-                      const std::string& base_path,
-                      Formula* formula);
+  /// @pre All the XML elements except for the last one are instructions.
+  ///
+  /// @post All forks in the event tree get registered.
+  ///
+  /// @throws ValidationError  Errors in instruction or target definitions.
+  void DefineBranch(const xmlpp::NodeSet& xml_nodes,
+                    EventTree* event_tree, Branch* branch);
 
-  /// Registers a basic event for later definition.
+  /// Processes Instruction definitions.
   ///
-  /// @param[in] event_node  XML element defining the event.
-  /// @param[in] base_path  Series of ancestor containers in the path with dots.
-  /// @param[in] container_role  The parent container's role.
+  /// @param[in] xml_element  The XML element with instruction definitions.
   ///
-  /// @returns Pointer to the registered basic event.
+  /// @returns The newly defined or registered instruction.
   ///
-  /// @throws ValidationError  An event with the same name is already defined.
-  BasicEventPtr RegisterBasicEvent(const xmlpp::Element* event_node,
-                                   const std::string& base_path,
-                                   RoleSpecifier container_role);
-
-  /// Defines a basic event for this analysis.
-  ///
-  /// @param[in] event_node  XML element defining the event.
-  /// @param[in,out] basic_event  Registered basic event ready to be defined.
-  void DefineBasicEvent(const xmlpp::Element* event_node,
-                        BasicEvent* basic_event);
-
-  /// Defines and adds a house event for this analysis.
-  ///
-  /// @param[in] event_node  XML element defining the event.
-  /// @param[in] base_path  Series of ancestor containers in the path with dots.
-  /// @param[in] container_role  The parent container's role.
-  ///
-  /// @returns Pointer to the registered house event.
-  ///
-  /// @throws ValidationError  An event with the same name is already defined.
-  HouseEventPtr DefineHouseEvent(const xmlpp::Element* event_node,
-                                 const std::string& base_path,
-                                 RoleSpecifier container_role);
-
-  /// Registers a variable or parameter.
-  ///
-  /// @param[in] param_node  XML element defining the parameter.
-  /// @param[in] base_path  Series of ancestor containers in the path with dots.
-  /// @param[in] container_role  The parent container's role.
-  ///
-  /// @returns Pointer to the registered parameter.
-  ///
-  /// @throws ValidationError  The parameter is already registered.
-  ParameterPtr RegisterParameter(const xmlpp::Element* param_node,
-                                 const std::string& base_path,
-                                 RoleSpecifier container_role);
-
-  /// Defines a variable or parameter.
-  ///
-  /// @param[in] param_node  XML element defining the parameter.
-  /// @param[in,out] parameter  Registered parameter to be defined.
-  void DefineParameter(const xmlpp::Element* param_node, Parameter* parameter);
+  /// @throws ValidationError  Errors in instruction definitions.
+  Instruction* GetInstruction(const xmlpp::Element* xml_element);
 
   /// Processes Expression definitions in input file.
   ///
   /// @param[in] expr_element  XML expression element containing the definition.
   /// @param[in] base_path  Series of ancestor containers in the path with dots.
   ///
-  /// @returns Pointer to the newly defined or registered expression.
+  /// @returns The newly defined or registered expression.
   ///
   /// @throws ValidationError  There are problems with getting the expression.
-  ExpressionPtr GetExpression(const xmlpp::Element* expr_element,
-                              const std::string& base_path);
-
-  /// Processes Constant Expression definitions in input file.
-  ///
-  /// @param[in] expr_element  XML expression element containing the definition.
-  ///
-  /// @returns Expression described in XML input expression node.
-  ExpressionPtr GetConstantExpression(const xmlpp::Element* expr_element);
+  Expression* GetExpression(const xmlpp::Element* expr_element,
+                            const std::string& base_path);
 
   /// Processes Parameter Expression definitions in input file.
   ///
+  /// @param[in] expr_type  The expression type name.
   /// @param[in] expr_element  XML expression element containing the definition.
   /// @param[in] base_path  Series of ancestor containers in the path with dots.
   ///
   /// @returns Parameter expression described in XML input expression node.
+  /// @returns nullptr if the expression type is not a parameter.
   ///
   /// @throws ValidationError  The parameter variable is not reachable.
-  ExpressionPtr GetParameterExpression(const xmlpp::Element* expr_element,
-                                       const std::string& base_path);
-
-  /// Registers a common cause failure group for later definition.
-  ///
-  /// @param[in] ccf_node  XML element defining CCF group.
-  /// @param[in] base_path  Series of ancestor containers in the path with dots.
-  /// @param[in] container_role  The parent container's role.
-  ///
-  /// @returns Pointer to the registered CCF group.
-  ///
-  /// @throws ValidationError  There are problems with registering
-  ///                          the group and its members,
-  ///                          for example, duplications or missing information.
-  CcfGroupPtr RegisterCcfGroup(const xmlpp::Element* ccf_node,
-                               const std::string& base_path,
-                               RoleSpecifier container_role);
-
-  /// Defines a common cause failure group for the analysis.
-  ///
-  /// @param[in] ccf_node  XML element defining CCF group.
-  /// @param[in,out] ccf_group  Registered CCF group to be defined.
-  void DefineCcfGroup(const xmlpp::Element* ccf_node, CcfGroup* ccf_group);
+  Expression* GetParameter(const std::string& expr_type,
+                           const xmlpp::Element* expr_element,
+                           const std::string& base_path);
 
   /// Processes common cause failure group members as defined basic events.
   ///
@@ -350,6 +332,49 @@ class Initializer : private boost::noncopyable {
   ///                          or factors for specific CCF models.
   void DefineCcfFactor(const xmlpp::Element* factor_node, CcfGroup* ccf_group);
 
+  /// Finds an entity (parameter, basic and house event, gate) from a reference.
+  /// The reference is case sensitive
+  /// and can contain an identifier, full path, or local path.
+  ///
+  /// @param[in] entity_reference  Reference string to the entity.
+  /// @param[in] base_path  The series of containers indicating the scope.
+  ///
+  /// @returns Pointer to the entity found by following the given reference.
+  ///
+  /// @throws std::out_of_range  The entity cannot be found.
+  /// @{
+  Parameter* GetParameter(const std::string& entity_reference,
+                          const std::string& base_path);
+  HouseEvent* GetHouseEvent(const std::string& entity_reference,
+                            const std::string& base_path);
+  BasicEvent* GetBasicEvent(const std::string& entity_reference,
+                            const std::string& base_path);
+  Gate* GetGate(const std::string& entity_reference,
+                const std::string& base_path);
+  Formula::EventArg GetEvent(const std::string& entity_reference,
+                             const std::string& base_path);
+  /// @}
+
+  /// Generic helper function to find an entity from a reference.
+  /// The reference is case sensitive
+  /// and can contain an identifier, full path, or local path.
+  ///
+  /// @tparam P  The pointer type managing the entity.
+  /// @tparam T  The entity type.
+  ///
+  /// @param[in] entity_reference  Reference string to the entity.
+  /// @param[in] base_path  The series of containers indicating the scope.
+  /// @param[in] container  Model's lookup container for entities with IDs.
+  /// @param[in] path_container  The full path container for entities.
+  ///
+  /// @returns Pointer to the requested entity.
+  ///
+  /// @throws std::out_of_range  The entity cannot be found.
+  template <class P, class T = typename P::element_type>
+  T* GetEntity(const std::string& entity_reference,
+               const std::string& base_path, const IdTable<P>& container,
+               const PathTable<T>& path_container);
+
   /// Validates if the initialization of the analysis is successful.
   ///
   /// @throws CycleError  Model contains cycles.
@@ -359,6 +384,34 @@ class Initializer : private boost::noncopyable {
   ///       if this error condition may lead resource leaks.
   void ValidateInitialization();
 
+  /// Checks the proper order of functional events in event tree forks.
+  ///
+  /// @param[in] branch  The event tree branch to start the check.
+  ///
+  /// @throws ValidationError  The order of forks is invalid.
+  ///
+  /// @pre All named branches are fed separately from initial states.
+  void CheckFunctionalEventOrder(const Branch& branch);
+
+  /// Checks that link instructions are used only in event-tree sequences.
+  ///
+  /// @param[in] branch  The event tree branch to start the check.
+  ///
+  /// @throws ValidationError  The Link instruction is misused.
+  ///
+  /// @pre All named branches are fed separately from initial states.
+  void EnsureLinksOnlyInSequences(const Branch& branch);
+
+  /// Ensures that event-tree does not mix
+  /// collect-expression and collect-formula.
+  ///
+  /// @param[in] branch  The event tree branch to start the check.
+  ///
+  /// @throws ValidationError  The Link instruction is misused.
+  ///
+  /// @pre All named branches are fed separately from initial states.
+  void EnsureHomogeneousEventTree(const Branch& branch);
+
   /// Validates expressions and anything
   /// that is dependent on them,
   /// such as parameters and basic events.
@@ -367,16 +420,6 @@ class Initializer : private boost::noncopyable {
   /// @throws ValidationError  There are problems detected with expressions.
   void ValidateExpressions();
 
-  /// Breaks all possible cycles in graph structures.
-  /// This function handles cycles
-  /// conservatively and indiscriminately.
-  ///
-  /// It may not be the most optimal approach,
-  /// but this error condition is considered uncommon.
-  ///
-  /// @post The model is unusable (freed).
-  void BreakCycles();
-
   /// Applies the input information to set up for future analysis.
   /// This step is crucial to get
   /// correct fault tree structures
@@ -388,29 +431,40 @@ class Initializer : private boost::noncopyable {
 
   std::shared_ptr<Model> model_;  ///< Analysis model with constructs.
   core::Settings settings_;  ///< Settings for analysis.
-  std::shared_ptr<MissionTime> mission_time_;  ///< Mission time expression.
 
   /// Saved parsers to keep XML documents alive.
   std::vector<std::unique_ptr<xmlpp::DomParser>> parsers_;
 
-  /// Map roots of documents to files. This is for error reporting.
+  /// Map roots of documents to files for error reporting.
   std::unordered_map<const xmlpp::Node*, std::string> doc_to_file_;
 
   /// Collection of elements that are defined late
   /// because of unordered registration and definition of their dependencies.
-  struct {
-    /// Parameters rely on parameter registration.
-    std::vector<std::pair<Parameter*, const xmlpp::Element*>> parameters;
-    /// Basic events rely on parameter registration.
-    std::vector<std::pair<BasicEvent*, const xmlpp::Element*>> basic_events;
-    /// Gates rely on gate, basic event, and house event registrations.
-    std::vector<std::pair<Gate*, const xmlpp::Element*>> gates;
-    /// CCF groups rely on both parameter and basic event registration.
-    std::vector<std::pair<CcfGroup*, const xmlpp::Element*>> ccf_groups;
-  } tbd_;  ///< Elements are assumed to be unique.
+  ///
+  /// Parameters and Expressions rely on parameter registrations.
+  /// Basic events rely on parameter registrations.
+  /// Gates rely on gate, basic event, and house event registrations.
+  /// CCF groups rely on both parameter and basic event registrations.
+  /// Event tree branches and instructions have complex interdependencies.
+  /// Initiating events may reference their associated event trees.
+  ///
+  /// Elements are assumed to be unique.
+  TbdContainer<Parameter, BasicEvent, Gate, CcfGroup, Sequence, EventTree,
+               InitiatingEvent, Rule>
+      tbd_;
 
-  /// Container for defined expressions for later validation.
-  std::vector<Expression*> expressions_;
+  /// Container of defined expressions for later validation due to cycles.
+  std::vector<std::pair<Expression*, const xmlpp::Element*>> expressions_;
+  /// Container for event tree links to check for cycles.
+  std::vector<Link*> links_;
+
+  /// Containers for reference resolution with paths.
+  /// @{
+  PathTable<Gate> path_gates_;
+  PathTable<BasicEvent> path_basic_events_;
+  PathTable<HouseEvent> path_house_events_;
+  PathTable<Parameter> path_parameters_;
+  /// @}
 };
 
 }  // namespace mef
