@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Olzhas Rakhimov
+ * Copyright (C) 2014-2017 Olzhas Rakhimov
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,11 +22,16 @@
 
 #include <fstream>
 #include <ostream>
+#include <utility>
 #include <vector>
 
+#include <boost/algorithm/string/join.hpp>
 #include <boost/date_time.hpp>
+#include <boost/range/adaptor/filtered.hpp>
+#include <boost/range/adaptor/transformed.hpp>
 
 #include "ccf_group.h"
+#include "element.h"
 #include "error.h"
 #include "logger.h"
 #include "parameter.h"
@@ -34,31 +39,56 @@
 
 namespace scram {
 
+namespace {
+
+/// Puts analysis id into report XML element.
+void PutId(const core::RiskAnalysis::Result::Id& id, XmlStreamElement* report) {
+  struct {
+    void operator()(const mef::Gate* gate) {
+      report_->SetAttribute("name", gate->id());
+    }
+    void operator()(const std::pair<const mef::InitiatingEvent&,
+                                    const mef::Sequence&>& sequence) {
+      report_->SetAttribute("initiating-event", sequence.first.name());
+      report_->SetAttribute("name", sequence.second.name());
+    }
+    XmlStreamElement* report_;
+  } extractor{report};
+  boost::apply_visitor(extractor, id);
+}
+
+}  // namespace
+
 void Reporter::Report(const core::RiskAnalysis& risk_an, std::ostream& out) {
   out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
   XmlStreamElement report("report", out);
   ReportInformation(risk_an, &report);
 
-  CLOCK(report_time);
-  LOG(DEBUG1) << "Reporting analysis results...";
+  if (risk_an.results().empty() && risk_an.event_tree_results().empty())
+    return;
+  TIMER(DEBUG1, "Reporting analysis results");
   XmlStreamElement results = report.AddChild("results");
-  for (const auto& fta : risk_an.fault_tree_analyses()) {
-    const std::string& id = fta.first;
-    core::ProbabilityAnalysis* prob_analysis = nullptr;  // Null if no analysis.
-    if (risk_an.settings().probability_analysis()) {
-      prob_analysis = risk_an.probability_analyses().at(id).get();
-    }
-    ReportResults(id, *fta.second, prob_analysis, &results);
-
-    if (risk_an.settings().importance_analysis()) {
-      ReportResults(id, *risk_an.importance_analyses().at(id), &results);
-    }
-
-    if (risk_an.settings().uncertainty_analysis()) {
-      ReportResults(id, *risk_an.uncertainty_analyses().at(id), &results);
+  if (risk_an.settings().probability_analysis()) {
+    for (const std::unique_ptr<core::EventTreeAnalysis>& result :
+         risk_an.event_tree_results()) {
+      ReportResults(*result, &results);
     }
   }
-  LOG(DEBUG1) << "Finished reporting in " << DUR(report_time);
+
+  for (const core::RiskAnalysis::Result& result : risk_an.results()) {
+    if (result.fault_tree_analysis)
+      ReportResults(result.id, *result.fault_tree_analysis,
+                    result.probability_analysis.get(), &results);
+
+    if (result.probability_analysis)
+      ReportResults(result.id, *result.probability_analysis, &results);
+
+    if (result.importance_analysis)
+      ReportResults(result.id, *result.importance_analysis, &results);
+
+    if (result.uncertainty_analysis)
+      ReportResults(result.id, *result.uncertainty_analysis, &results);
+  }
 }
 
 void Reporter::Report(const core::RiskAnalysis& risk_an,
@@ -82,16 +112,16 @@ void Reporter::ReportCalculatedQuantity<core::FaultTreeAnalysis>(
     } else {
       quant.SetAttribute("name", "Minimal Cut Sets");
     }
-  }
-  {
-    XmlStreamElement methods = information->AddChild("calculation-method");
-    if (settings.algorithm() == "bdd") {
-      methods.SetAttribute("name", "Binary Decision Diagram");
-    } else if (settings.algorithm() == "zbdd") {
-      methods.SetAttribute("name", "Zero-Suppressed Binary Decision Diagram");
-    } else {
-      assert(settings.algorithm() == "mocus");
-      methods.SetAttribute("name", "MOCUS");
+    XmlStreamElement methods = quant.AddChild("calculation-method");
+    switch (settings.algorithm()) {
+      case core::Algorithm::kBdd:
+        methods.SetAttribute("name", "Binary Decision Diagram");
+        break;
+      case core::Algorithm::kZbdd:
+        methods.SetAttribute("name", "Zero-Suppressed Binary Decision Diagram");
+        break;
+      case core::Algorithm::kMocus:
+        methods.SetAttribute("name", "MOCUS");
     }
     methods.AddChild("limits")
         .AddChild("product-order")
@@ -110,23 +140,30 @@ template <>
 void Reporter::ReportCalculatedQuantity<core::ProbabilityAnalysis>(
     const core::Settings& settings,
     XmlStreamElement* information) {
-  information->AddChild("calculated-quantity")
-      .SetAttribute("name", "Probability Analysis")
+  XmlStreamElement quant = information->AddChild("calculated-quantity");
+  quant.SetAttribute("name", "Probability Analysis")
       .SetAttribute("definition",
-                    "Quantitative analysis of failure probability")
-      .SetAttribute("approximation", settings.approximation());
+                    "Quantitative analysis of"
+                    " failure probability or unavailability")
+      .SetAttribute("approximation",
+                    core::kApproximationToString[static_cast<int>(
+                        settings.approximation())]);
 
-  XmlStreamElement methods = information->AddChild("calculation-method");
-  if (settings.approximation() == "rare-event") {
-    methods.SetAttribute("name", "Rare-Event Approximation");
-  } else if (settings.approximation() == "mcub") {
-    methods.SetAttribute("name", "MCUB Approximation");
-  } else {
-    assert(settings.approximation() == "no");
-    methods.SetAttribute("name", "Binary Decision Diagram");
+  XmlStreamElement methods = quant.AddChild("calculation-method");
+  switch (settings.approximation()) {
+    case core::Approximation::kNone:
+      methods.SetAttribute("name", "Binary Decision Diagram");
+      break;
+    case core::Approximation::kRareEvent:
+      methods.SetAttribute("name", "Rare-Event Approximation");
+      break;
+    case core::Approximation::kMcub:
+      methods.SetAttribute("name", "MCUB Approximation");
   }
   XmlStreamElement limits = methods.AddChild("limits");
   limits.AddChild("mission-time").AddText(settings.mission_time());
+  if (settings.time_step())
+    limits.AddChild("time-step").AddText(settings.time_step());
 }
 
 /// Describes the importance analysis and techniques.
@@ -146,12 +183,12 @@ template <>
 void Reporter::ReportCalculatedQuantity<core::UncertaintyAnalysis>(
     const core::Settings& settings,
     XmlStreamElement* information) {
-  information->AddChild("calculated-quantity")
-      .SetAttribute("name", "Uncertainty Analysis")
+  XmlStreamElement quant = information->AddChild("calculated-quantity");
+  quant.SetAttribute("name", "Uncertainty Analysis")
       .SetAttribute("definition",
                     "Calculation of uncertainties with the Monte Carlo method");
 
-  XmlStreamElement methods = information->AddChild("calculation-method");
+  XmlStreamElement methods = quant.AddChild("calculation-method");
   methods.SetAttribute("name", "Monte Carlo");
   XmlStreamElement limits = methods.AddChild("limits");
   limits.AddChild("number-of-trials").AddText(settings.num_trials());
@@ -171,6 +208,10 @@ void Reporter::ReportCalculatedQuantity<core::RiskAnalysis>(
   if (settings.probability_analysis()) {
     ReportCalculatedQuantity<core::ProbabilityAnalysis>(settings, information);
   }
+  if (settings.safety_integrity_levels()) {
+    information->AddChild("calculated-quantity")
+        .SetAttribute("name", "Safety Integrity Levels");
+  }
   if (settings.importance_analysis()) {
     ReportCalculatedQuantity<core::ImportanceAnalysis>(settings, information);
   }
@@ -186,14 +227,36 @@ void Reporter::ReportInformation(const core::RiskAnalysis& risk_an,
   ReportPerformance(risk_an, &information);
   ReportCalculatedQuantity(risk_an.settings(), &information);
   ReportModelFeatures(risk_an.model(), &information);
-  ReportOrphanPrimaryEvents(risk_an.model(), &information);
-  ReportUnusedParameters(risk_an.model(), &information);
+  ReportUnusedElements(risk_an.model().basic_events(), "Unused basic events: ",
+                       &information);
+  ReportUnusedElements(risk_an.model().house_events(), "Unused house events: ",
+                       &information);
+  ReportUnusedElements(risk_an.model().parameters(), "Unused parameters: ",
+                       &information);
+  ReportUnusedElements(risk_an.model().initiating_events(),
+                       "Unused initiating events: ", &information);
+  ReportUnusedElements(risk_an.model().event_trees(),
+                       "Unused event trees: ", &information);
+  ReportUnusedElements(risk_an.model().sequences(), "Unused sequences: ",
+                       &information);
+  ReportUnusedElements(risk_an.model().rules(), "Unused rules: ",
+                       &information);
+  for (const mef::EventTreePtr& event_tree : risk_an.model().event_trees()) {
+    std::string header = "In event tree " + event_tree->name() + ", ";
+    ReportUnusedElements(event_tree->branches(), header + "unused branches: ",
+                         &information);
+    ReportUnusedElements(event_tree->functional_events(),
+                         header + "unused functional events: ", &information);
+  }
 }
 
 void Reporter::ReportSoftwareInformation(XmlStreamElement* information) {
   information->AddChild("software")
       .SetAttribute("name", "SCRAM")
-      .SetAttribute("version", version::core());
+      .SetAttribute("version", *version::describe() != '\0'
+                                   ? version::describe()
+                                   : version::core())
+      .SetAttribute("contacts", "https://scram-pra.org");
   namespace pt = boost::posix_time;
   information->AddChild("time").AddText(
       pt::to_iso_extended_string(pt::second_clock::universal_time()));
@@ -202,146 +265,212 @@ void Reporter::ReportSoftwareInformation(XmlStreamElement* information) {
 void Reporter::ReportModelFeatures(const mef::Model& model,
                                    XmlStreamElement* information) {
   XmlStreamElement model_features = information->AddChild("model-features");
-  if (!model.name().empty())
+  if (!model.HasDefaultName())
     model_features.SetAttribute("name", model.name());
-  model_features.AddChild("gates").AddText(model.gates().size());
-  model_features.AddChild("basic-events").AddText(model.basic_events().size());
-  model_features.AddChild("house-events").AddText(model.house_events().size());
-  model_features.AddChild("ccf-groups").AddText(model.ccf_groups().size());
-  model_features.AddChild("fault-trees").AddText(model.fault_trees().size());
+  auto feature = [&model_features](const char* name, const auto& container) {
+    if (!container.empty())
+      model_features.AddChild(name).AddText(container.size());
+  };
+  feature("gates", model.gates());
+  feature("basic-events", model.basic_events());
+  feature("house-events", model.house_events());
+  feature("ccf-groups", model.ccf_groups());
+  feature("fault-trees", model.fault_trees());
+  feature("event-trees", model.event_trees());
+
+  int num_functional_events = 0;
+  for (const mef::EventTreePtr& event_tree : model.event_trees())
+    num_functional_events += event_tree->functional_events().size();
+  if (num_functional_events)
+    model_features.AddChild("functional-events").AddText(num_functional_events);
+
+  feature("sequences", model.sequences());
+  feature("rules", model.rules());
+  feature("initiating-events", model.initiating_events());
 }
 
 void Reporter::ReportPerformance(const core::RiskAnalysis& risk_an,
                                  XmlStreamElement* information) {
+  if (risk_an.results().empty())
+    return;
   // Setup for performance information.
   XmlStreamElement performance = information->AddChild("performance");
-  for (const auto& fta : risk_an.fault_tree_analyses()) {
+  for (const core::RiskAnalysis::Result& result : risk_an.results()) {
     XmlStreamElement calc_time = performance.AddChild("calculation-time");
-    const std::string& id = fta.first;
-    calc_time.SetAttribute("name", id);
-    calc_time.AddChild("products").AddText(fta.second->analysis_time());
+    scram::PutId(result.id, &calc_time);
+    if (result.fault_tree_analysis)
+      calc_time.AddChild("products")
+          .AddText(result.fault_tree_analysis->analysis_time());
 
-    if (risk_an.settings().probability_analysis()) {
+    if (result.probability_analysis)
       calc_time.AddChild("probability")
-          .AddText(risk_an.probability_analyses().at(id)->analysis_time());
-    }
+          .AddText(result.probability_analysis->analysis_time());
 
-    if (risk_an.settings().importance_analysis()) {
+    if (result.importance_analysis)
       calc_time.AddChild("importance")
-          .AddText(risk_an.importance_analyses().at(id)->analysis_time());
-    }
+          .AddText(result.importance_analysis->analysis_time());
 
-    if (risk_an.settings().uncertainty_analysis()) {
+    if (result.uncertainty_analysis)
       calc_time.AddChild("uncertainty")
-          .AddText(risk_an.uncertainty_analyses().at(id)->analysis_time());
-    }
+          .AddText(result.uncertainty_analysis->analysis_time());
   }
 }
 
-void Reporter::ReportOrphanPrimaryEvents(const mef::Model& model,
-                                         XmlStreamElement* information) {
-  std::string out;
-  for (const mef::BasicEventPtr& param : model.basic_events()) {
-    if (param->orphan())
-      out += param->id() + " ";
-  }
-  for (const mef::HouseEventPtr& param : model.house_events()) {
-    if (param->orphan())
-      out += param->id() + " ";
-  }
+template <class T>
+void Reporter::ReportUnusedElements(const T& container,
+                                    const std::string& header,
+                                    XmlStreamElement* information) {
+  std::string out = boost::join(
+      container | boost::adaptors::filtered([](auto& ptr) {
+        return !ptr->usage();
+      }) | boost::adaptors::transformed([](auto& ptr) -> decltype(auto) {
+        return mef::Id::unique_name(*ptr);
+      }),
+      " ");
   if (!out.empty())
-    information->AddChild("warning").AddText("Orphan Primary Events: " + out);
+    information->AddChild("warning").AddText(header + out);
 }
 
-void Reporter::ReportUnusedParameters(const mef::Model& model,
-                                      XmlStreamElement* information) {
-  std::string out;
-  for (const mef::ParameterPtr& param : model.parameters()) {
-    if (param->unused())
-      out += param->id() + " ";
+void Reporter::ReportResults(const core::EventTreeAnalysis& eta,
+                             XmlStreamElement* results) {
+  XmlStreamElement initiating_event = results->AddChild("initiating-event");
+  initiating_event.SetAttribute("name", eta.initiating_event().name())
+      .SetAttribute("sequences", eta.sequences().size());
+  for (const core::EventTreeAnalysis::Result& result_sequence :
+       eta.sequences()) {
+    initiating_event.AddChild("sequence")
+        .SetAttribute("name", result_sequence.sequence.name())
+        .SetAttribute("value", result_sequence.p_sequence);
   }
-  if (!out.empty())
-    information->AddChild("warning").AddText("Unused Parameters: " + out);
 }
 
-void Reporter::ReportResults(const std::string& ft_name,
+void Reporter::ReportResults(const core::RiskAnalysis::Result::Id& id,
                              const core::FaultTreeAnalysis& fta,
                              const core::ProbabilityAnalysis* prob_analysis,
                              XmlStreamElement* results) {
+  TIMER(DEBUG2, "Reporting products");
   XmlStreamElement sum_of_products = results->AddChild("sum-of-products");
-  sum_of_products.SetAttribute("name", ft_name)
-      .SetAttribute("basic-events", fta.product_events().size())
-      .SetAttribute("products", fta.products().size());
+  scram::PutId(id, &sum_of_products);
 
   std::string warning = fta.warnings();
-  if (prob_analysis) {
+  if (prob_analysis && prob_analysis->warnings().empty() == false)
+    warning += (warning.empty() ? "" : "; ") + prob_analysis->warnings();
+  if (!warning.empty())
+    sum_of_products.SetAttribute("warning", warning);
+
+  sum_of_products
+      .SetAttribute("basic-events", fta.products().product_events().size())
+      .SetAttribute("products", fta.products().size());
+
+  if (prob_analysis)
     sum_of_products.SetAttribute("probability", prob_analysis->p_total());
-    warning += prob_analysis->warnings();
-  }
-  if (!warning.empty()) {
-    sum_of_products.AddChild("warning").AddText(warning);
+
+  if (fta.products().empty() == false) {
+    sum_of_products.SetAttribute(
+        "distribution",
+        boost::join(fta.products().Distribution() |
+                        boost::adaptors::transformed(
+                            [](int number) { return std::to_string(number); }),
+                    " "));
   }
 
-  CLOCK(cs_time);
-  LOG(DEBUG2) << "Reporting products for " << ft_name << "...";
-  std::vector<double> probs;  // Product probabilities.
   double sum = 0;  // Sum of probabilities for contribution calculations.
   if (prob_analysis) {
-    for (const core::Product& product_set : fta.products()) {
-      double prob = CalculateProbability(product_set);
-      sum += prob;
-      probs.push_back(prob);
-    }
-  }  // Ugliness because FTA and Probability analyses are not integrated.
-  for (int i = 0; i < fta.products().size(); ++i) {
-    const core::Product& product_set = fta.products()[i];
+    for (const core::Product& product_set : fta.products())
+      sum += product_set.p();
+  }
+  for (const core::Product& product_set : fta.products()) {
     XmlStreamElement product = sum_of_products.AddChild("product");
-    product.SetAttribute("order", GetOrder(product_set));
+    product.SetAttribute("order", product_set.order());
     if (prob_analysis) {
-      double prob = probs[i];
+      double prob = product_set.p();
       product.SetAttribute("probability", prob);
-      product.SetAttribute("contribution", prob / sum);
+      if (sum != 0)
+        product.SetAttribute("contribution", prob / sum);
     }
     for (const core::Literal& literal : product_set) {
       ReportLiteral(literal, &product);
     }
   }
-  LOG(DEBUG2) << "Finished reporting products in " << DUR(cs_time);
+}
+
+void Reporter::ReportResults(const core::RiskAnalysis::Result::Id& id,
+                             const core::ProbabilityAnalysis& prob_analysis,
+                             XmlStreamElement* results) {
+  if (!prob_analysis.p_time().empty()) {
+    XmlStreamElement curve = results->AddChild("curve");
+    scram::PutId(id, &curve);
+    curve.SetAttribute("description", "Probability values over time")
+        .SetAttribute("X-title", "Mission time")
+        .SetAttribute("Y-title", "Probability")
+        .SetAttribute("X-unit", "hours");
+    for (const std::pair<double, double>& p_vs_time : prob_analysis.p_time()) {
+      curve.AddChild("point")
+          .SetAttribute("X", p_vs_time.second)
+          .SetAttribute("Y", p_vs_time.first);
+    }
+  }
+
+  if (prob_analysis.settings().safety_integrity_levels()) {
+    XmlStreamElement sil = results->AddChild("safety-integrity-levels");
+    scram::PutId(id, &sil);
+    sil.SetAttribute("PFD-avg", prob_analysis.sil().pfd_avg)
+        .SetAttribute("PFH-avg", prob_analysis.sil().pfh_avg);
+    auto report_sil_fractions = [&sil](const auto& sil_fractions) {
+      XmlStreamElement hist = sil.AddChild("histogram");
+      hist.SetAttribute("number", sil_fractions.size());
+      double b_0 = 0;
+      int bin_number = 1;
+      for (const std::pair<const double, double>& sil_bucket : sil_fractions) {
+        double b_1 = sil_bucket.first;
+        hist.AddChild("bin")
+            .SetAttribute("number", bin_number++)
+            .SetAttribute("value", sil_bucket.second)
+            .SetAttribute("lower-bound", b_0)
+            .SetAttribute("upper-bound", b_1);
+        b_0 = b_1;
+      }
+    };
+    report_sil_fractions(prob_analysis.sil().pfd_fractions);
+    report_sil_fractions(prob_analysis.sil().pfh_fractions);
+  }
 }
 
 void Reporter::ReportResults(
-    const std::string& ft_name,
+    const core::RiskAnalysis::Result::Id& id,
     const core::ImportanceAnalysis& importance_analysis,
     XmlStreamElement* results) {
   XmlStreamElement importance = results->AddChild("importance");
-  importance.SetAttribute("name", ft_name);
+  scram::PutId(id, &importance);
+  if (!importance_analysis.warnings().empty()) {
+    importance.SetAttribute("warning", importance_analysis.warnings());
+  }
   importance.SetAttribute("basic-events",
                           importance_analysis.importance().size());
-  if (!importance_analysis.warnings().empty()) {
-    importance.AddChild("warning").AddText(importance_analysis.warnings());
-  }
 
-  for (const core::ImportanceAnalysis::ImportanceRecord& entry :
-       importance_analysis.important_events()) {
-    const core::ImportanceFactors& factors = entry.second;
-    auto add_data = [&factors](XmlStreamElement* element) {
-      element->SetAttribute("MIF", factors.mif)
+  for (const core::ImportanceRecord& entry : importance_analysis.importance()) {
+    const core::ImportanceFactors& factors = entry.factors;
+    const mef::BasicEvent& event = entry.event;
+    auto add_data = [&event, &factors](XmlStreamElement* element) {
+      element->SetAttribute("occurrence", factors.occurrence)
+          .SetAttribute("probability", event.p())
+          .SetAttribute("MIF", factors.mif)
           .SetAttribute("CIF", factors.cif)
           .SetAttribute("DIF", factors.dif)
           .SetAttribute("RAW", factors.raw)
           .SetAttribute("RRW", factors.rrw);
     };
-    ReportBasicEvent(*entry.first, &importance, add_data);
+    ReportBasicEvent(event, &importance, add_data);
   }
 }
 
-void Reporter::ReportResults(const std::string& ft_name,
+void Reporter::ReportResults(const core::RiskAnalysis::Result::Id& id,
                              const core::UncertaintyAnalysis& uncert_analysis,
                              XmlStreamElement* results) {
   XmlStreamElement measure = results->AddChild("measure");
-  measure.SetAttribute("name", ft_name);
+  scram::PutId(id, &measure);
   if (!uncert_analysis.warnings().empty()) {
-    measure.AddChild("warning").AddText(uncert_analysis.warnings());
+    measure.SetAttribute("warning", uncert_analysis.warnings());
   }
   measure.AddChild("mean").SetAttribute("value", uncert_analysis.mean());
   measure.AddChild("standard-deviation")
