@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Olzhas Rakhimov
+ * Copyright (C) 2014-2017 Olzhas Rakhimov
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -50,7 +50,7 @@ void RiskAnalysisTest::ProcessInputFiles(
     const std::vector<std::string>& input_files) {
   mef::Initializer init(input_files, settings);
   model = init.model();
-  analysis = std::make_unique<RiskAnalysis>(model, settings);
+  analysis = std::make_unique<RiskAnalysis>(model.get(), settings);
   result_ = Result();
 }
 
@@ -68,12 +68,10 @@ void RiskAnalysisTest::CheckReport(const std::string& tree_input) {
 }
 
 const std::set<std::set<std::string>>& RiskAnalysisTest::products() {
-  assert(!analysis->fault_tree_analyses().empty());
-  assert(analysis->fault_tree_analyses().size() == 1);
+  assert(analysis->results().size() == 1);
   if (result_.products.empty()) {
-    const FaultTreeAnalysis* fta =
-        analysis->fault_tree_analyses().begin()->second.get();
-    for (const Product& product : fta->products()) {
+    for (const Product& product :
+         analysis->results().front().fault_tree_analysis->products()) {
       result_.products.emplace(Convert(product));
     }
   }
@@ -81,42 +79,40 @@ const std::set<std::set<std::string>>& RiskAnalysisTest::products() {
 }
 
 std::vector<int> RiskAnalysisTest::ProductDistribution() {
-  assert(!analysis->fault_tree_analyses().empty());
-  assert(analysis->fault_tree_analyses().size() == 1);
-  std::vector<int> distr(settings.limit_order(), 0);
-  const FaultTreeAnalysis* fta =
-      analysis->fault_tree_analyses().begin()->second.get();
-  for (const Product& product : fta->products()) {
-    distr[GetOrder(product) - 1]++;
-  }
-  while (!distr.empty() && !distr.back())
-    distr.pop_back();
-  return distr;
+  assert(analysis->results().size() == 1);
+  return analysis->results()
+      .front()
+      .fault_tree_analysis->products()
+      .Distribution();
 }
 
 void RiskAnalysisTest::PrintProducts() {
-  assert(!analysis->fault_tree_analyses().empty());
-  assert(analysis->fault_tree_analyses().size() == 1);
-  const FaultTreeAnalysis* fta =
-      analysis->fault_tree_analyses().begin()->second.get();
-  Print(fta->products());
+  assert(analysis->results().size() == 1);
+  Print(analysis->results().front().fault_tree_analysis->products());
 }
 
 const std::map<std::set<std::string>, double>&
 RiskAnalysisTest::product_probability() {
-  assert(!analysis->fault_tree_analyses().empty());
-  assert(analysis->fault_tree_analyses().size() == 1);
+  assert(analysis->results().size() == 1);
   if (result_.product_probability.empty()) {
-    const FaultTreeAnalysis* fta =
-        analysis->fault_tree_analyses().begin()->second.get();
-    for (const Product& product : fta->products()) {
-      result_.product_probability.emplace(Convert(product),
-                                          CalculateProbability(product));
+    for (const Product& product :
+         analysis->results().front().fault_tree_analysis->products()) {
+      result_.product_probability.emplace(Convert(product), product.p());
     }
   }
   return result_.product_probability;
 }
 
+std::map<std::string, double> RiskAnalysisTest::sequences() {
+  assert(analysis->event_tree_results().size() == 1);
+  std::map<std::string, double> results;
+  for (const core::EventTreeAnalysis::Result& result :
+       analysis->event_tree_results().front()->sequences()) {
+    results.emplace(result.sequence.name(), result.p_sequence);
+  }
+  return results;
+}
+
 std::set<std::string> RiskAnalysisTest::Convert(const Product& product) {
   std::set<std::string> string_set;
   for (const Literal& literal : product) {
@@ -139,21 +135,21 @@ TEST_F(RiskAnalysisTest, ProcessInput) {
   EXPECT_EQ(1, basic_events().count("ValveTwo"));
 
   ASSERT_TRUE(gates().count("TopEvent"));
-  mef::GatePtr top = *gates().find("TopEvent");
+  mef::Gate* top = gates().find("TopEvent")->get();
   EXPECT_EQ("TopEvent", top->id());
   ASSERT_NO_THROW(top->formula().type());
   EXPECT_EQ(mef::kAnd, top->formula().type());
   EXPECT_EQ(2, top->formula().event_args().size());
 
   ASSERT_TRUE(gates().count("TrainOne"));
-  mef::GatePtr inter = *gates().find("TrainOne");
+  mef::Gate* inter = gates().find("TrainOne")->get();
   EXPECT_EQ("TrainOne", inter->id());
   ASSERT_NO_THROW(inter->formula().type());
   EXPECT_EQ(mef::kOr, inter->formula().type());
   EXPECT_EQ(2, inter->formula().event_args().size());
 
   ASSERT_TRUE(basic_events().count("ValveOne"));
-  mef::BasicEventPtr primary = *basic_events().find("ValveOne");
+  mef::BasicEvent* primary = basic_events().find("ValveOne")->get();
   EXPECT_EQ("ValveOne", primary->id());
 }
 
@@ -169,10 +165,10 @@ TEST_F(RiskAnalysisTest, PopulateProbabilities) {
   ASSERT_EQ(1, basic_events().count("ValveOne"));
   ASSERT_EQ(1, basic_events().count("ValveTwo"));
 
-  mef::BasicEventPtr p1 = *basic_events().find("PumpOne");
-  mef::BasicEventPtr p2 = *basic_events().find("PumpTwo");
-  mef::BasicEventPtr v1 = *basic_events().find("ValveOne");
-  mef::BasicEventPtr v2 = *basic_events().find("ValveTwo");
+  mef::BasicEvent* p1 = basic_events().find("PumpOne")->get();
+  mef::BasicEvent* p2 = basic_events().find("PumpTwo")->get();
+  mef::BasicEvent* v1 = basic_events().find("ValveOne")->get();
+  mef::BasicEvent* v2 = basic_events().find("ValveTwo")->get();
 
   ASSERT_NO_THROW(p1->p());
   ASSERT_NO_THROW(p2->p());
@@ -230,7 +226,7 @@ TEST_P(RiskAnalysisTest, AnalyzeWithProbability) {
   ASSERT_NO_THROW(analysis->Analyze());
 
   EXPECT_EQ(mcs, products());
-  if (settings.approximation() == "rare-event") {
+  if (settings.approximation() == Approximation::kRareEvent) {
     EXPECT_DOUBLE_EQ(1, p_total());
   } else {
     EXPECT_DOUBLE_EQ(0.646, p_total());
@@ -246,7 +242,7 @@ TEST_P(RiskAnalysisTest, AnalyzeWithProbability) {
 TEST_P(RiskAnalysisTest, EnforceExactProbability) {
   std::string with_prob =
       "./share/scram/input/fta/correct_tree_input_with_probs.xml";
-  settings.probability_analysis(true).approximation("no");
+  settings.probability_analysis(true).approximation("none");
   ASSERT_NO_THROW(ProcessInputFile(with_prob));
   ASSERT_NO_THROW(analysis->Analyze());
   EXPECT_DOUBLE_EQ(0.646, p_total());
@@ -269,22 +265,10 @@ TEST_F(RiskAnalysisTest, ImportanceDefault) {
   settings.importance_analysis(true);
   ASSERT_NO_THROW(ProcessInputFile(with_prob));
   ASSERT_NO_THROW(analysis->Analyze());
-  // Check importance values.
-  std::vector<std::pair<std::string, ImportanceFactors>> expected_results = {
-      {"PumpOne", {0.51, 0.4737, 0.7895, 1.316, 1.9}},
-      {"PumpTwo", {0.38, 0.4118, 0.8235, 1.176, 1.7}},
-      {"ValveOne", {0.34, 0.2105, 0.5263, 1.316, 1.267}},
-      {"ValveTwo", {0.228, 0.1765, 0.5882, 1.176, 1.214}}};
-
-  for (const auto& entry : expected_results) {
-    const ImportanceFactors& result = importance(entry.first);
-    const ImportanceFactors& test = entry.second;
-    EXPECT_NEAR(test.mif, result.mif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.cif, result.cif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.dif, result.dif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.raw, result.raw, 1e-3) << entry.first;
-    EXPECT_NEAR(test.rrw, result.rrw, 1e-3) << entry.first;
-  }
+  TestImportance({{"PumpOne", {2, 0.51, 0.4737, 0.7895, 1.316, 1.9}},
+                  {"PumpTwo", {2, 0.38, 0.4118, 0.8235, 1.176, 1.7}},
+                  {"ValveOne", {2, 0.34, 0.2105, 0.5263, 1.316, 1.267}},
+                  {"ValveTwo", {2, 0.228, 0.1765, 0.5882, 1.176, 1.214}}});
 }
 
 TEST_F(RiskAnalysisTest, ImportanceNeg) {
@@ -294,21 +278,34 @@ TEST_F(RiskAnalysisTest, ImportanceNeg) {
   ASSERT_NO_THROW(analysis->Analyze());
   EXPECT_NEAR(0.04459, p_total(), 1e-3);
   // Check importance values with negative event.
-  std::vector<std::pair<std::string, ImportanceFactors>> expected_results = {
-      {"PumpOne", {0.0765, 0.1029, 0.1568, 2.613, 1.115}},
-      {"PumpTwo", {0.057, 0.08948, 0.1532, 2.189, 1.098}},
-      {"ValveOne", {0.94, 0.8432, 0.8495, 21.237, 6.379}},
-      {"ValveTwo", {0.0558, 0.06257, 0.1094, 2.189, 1.067}}};
+  TestImportance({{"PumpOne", {3, 0.0765, 0.1029, 0.1568, 2.613, 1.115}},
+                  {"PumpTwo", {2, 0.057, 0.08948, 0.1532, 2.189, 1.098}},
+                  {"ValveOne", {3, 0.94, 0.8432, 0.8495, 21.237, 6.379}},
+                  {"ValveTwo", {2, 0.0558, 0.06257, 0.1094, 2.189, 1.067}}});
+}
 
-  for (const auto& entry : expected_results) {
-    const ImportanceFactors& result = importance(entry.first);
-    const ImportanceFactors& test = entry.second;
-    EXPECT_NEAR(test.mif, result.mif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.cif, result.cif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.dif, result.dif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.raw, result.raw, 1e-3) << entry.first;
-    EXPECT_NEAR(test.rrw, result.rrw, 1e-3) << entry.first;
-  }
+TEST_P(RiskAnalysisTest, ImportanceSingleEvent) {
+  std::string tree_input = "./share/scram/input/core/null_a.xml";
+  settings.importance_analysis(true);
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  TestImportance({{"OnlyChild", {1, 1, 1, 1, 2, 0}}});
+}
+
+TEST_P(RiskAnalysisTest, ImportanceZeroProbability) {
+  std::string tree_input = "./share/scram/input/core/zero_prob.xml";
+  settings.importance_analysis(true);
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  TestImportance({{"A", {1, 1, 0, 0, 0, 0}}});
+}
+
+TEST_P(RiskAnalysisTest, ImportanceOneProbability) {
+  std::string tree_input = "./share/scram/input/core/one_prob.xml";
+  settings.importance_analysis(true);
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  TestImportance({{"A", {1, 1, 1, 1, 1, 0}}});
 }
 
 // Apply the rare event approximation.
@@ -319,22 +316,10 @@ TEST_F(RiskAnalysisTest, ImportanceRareEvent) {
   ASSERT_NO_THROW(ProcessInputFile(with_prob));
   ASSERT_NO_THROW(analysis->Analyze());
   EXPECT_DOUBLE_EQ(0.012, p_total());  // Adjusted probability.
-  // Check importance values.
-  std::vector<std::pair<std::string, ImportanceFactors>> expected_results = {
-      {"PumpOne", {0.12, 0.6, 0.624, 10.4, 2.5}},
-      {"PumpTwo", {0.1, 0.5833, 0.6125, 8.75, 2.4}},
-      {"ValveOne", {0.12, 0.4, 0.424, 10.6, 1.667}},
-      {"ValveTwo", {0.1, 0.4167, 0.4458, 8.917, 1.714}}};
-
-  for (const auto& entry : expected_results) {
-    const ImportanceFactors& result = importance(entry.first);
-    const ImportanceFactors& test = entry.second;
-    EXPECT_NEAR(test.mif, result.mif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.cif, result.cif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.dif, result.dif, 1e-3) << entry.first;
-    EXPECT_NEAR(test.raw, result.raw, 1e-3) << entry.first;
-    EXPECT_NEAR(test.rrw, result.rrw, 1e-3) << entry.first;
-  }
+  TestImportance({{"PumpOne", {2, 0.12, 0.6, 0.624, 10.4, 2.5}},
+                  {"PumpTwo", {2, 0.1, 0.5833, 0.6125, 8.75, 2.4}},
+                  {"ValveOne", {2, 0.12, 0.4, 0.424, 10.6, 1.667}},
+                  {"ValveTwo", {2, 0.1, 0.4167, 0.4458, 8.917, 1.714}}});
 }
 
 // Apply the minimal cut set upper bound approximation.
@@ -369,6 +354,116 @@ TEST_P(RiskAnalysisTest, AnalyzeMC) {
   ASSERT_NO_THROW(analysis->Analyze());
 }
 
+TEST_P(RiskAnalysisTest, AnalyzeProbabilityOverTime) {
+  std::string tree_input = "./share/scram/input/core/single_exponential.xml";
+  settings.probability_analysis(true).time_step(24).mission_time(120);
+  std::vector<double> curve = {0,        2.399e-4, 4.7989e-4,
+                               7.197e-4, 9.595e-4, 1.199e-3};
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  ASSERT_FALSE(analysis->results().empty());
+  ASSERT_TRUE(analysis->results().front().probability_analysis);
+  auto it = curve.begin();
+  double time = 0;
+  for (const std::pair<double, double>& p_vs_time :
+       analysis->results().front().probability_analysis->p_time()) {
+    ASSERT_NE(curve.end(), it);
+    if (time >= settings.mission_time()) {
+      EXPECT_EQ(settings.mission_time(), p_vs_time.second);
+    } else {
+      EXPECT_EQ(time, p_vs_time.second);
+    }
+    EXPECT_NEAR(*it, p_vs_time.first, *it * 0.001);
+    time += settings.time_step();
+    ++it;
+  }
+  ASSERT_TRUE(time);
+}
+
+TEST_P(RiskAnalysisTest, AnalyzeSil) {
+  std::string tree_input = "./share/scram/input/core/single_exponential.xml";
+  settings.time_step(24).safety_integrity_levels(true);
+  double pfd_fractions[] = {1.142e-4, 1.0275e-3, 1.02796e-2,
+                            0.1033,   0.88527,   0};
+  double pfh_fractions[] = {2.74e-7, 2.466e-6, 2.466e-5, 2.466e-4, 0.999726, 0};
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  ASSERT_FALSE(analysis->results().empty());
+  ASSERT_TRUE(analysis->results().front().probability_analysis);
+  const auto& prob_an = *analysis->results().front().probability_analysis;
+  EXPECT_NEAR(0.04255, prob_an.sil().pfd_avg, 0.00001);
+  EXPECT_NEAR(9.77e-6, prob_an.sil().pfh_avg, 1e-8);
+  auto compare_fractions = [](const auto& sil_fractions, const auto& result,
+                              const char* type) {
+    auto it = std::begin(sil_fractions);
+    for (const std::pair<const double, double>& result_bucket : result) {
+      ASSERT_NE(std::end(sil_fractions), it);
+      EXPECT_NEAR(*it, result_bucket.second, *it * 0.001)
+          << "The " << type << " bucket for " << result_bucket.first;
+      ++it;
+    }
+    ASSERT_EQ(std::end(sil_fractions), it);
+  };
+  compare_fractions(pfd_fractions, prob_an.sil().pfd_fractions, "PFD");
+  compare_fractions(pfh_fractions, prob_an.sil().pfh_fractions, "PFH");
+}
+
+TEST_P(RiskAnalysisTest, AnalyzeEventTree) {
+  const char* tree_input = "./share/scram/input/EventTrees/bcd.xml";
+  settings.probability_analysis(true);
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  EXPECT_EQ(1, analysis->event_tree_results().size());
+  const auto& results = sequences();
+  ASSERT_EQ(2, results.size());
+  std::map<std::string, double> expected = {{"Success", 0.594},
+                                            {"Failure", 0.406}};
+  for (const auto& result : expected) {
+    ASSERT_TRUE(results.count(result.first)) << result.first;
+    EXPECT_DOUBLE_EQ(result.second, results.at(result.first)) << result.first;
+  }
+}
+
+TEST_P(RiskAnalysisTest, AnalyzeTestEventDefault) {
+  const char* tree_input = "./share/scram/input/eta/test_event_default.xml";
+  settings.probability_analysis(true);
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  EXPECT_EQ(1, analysis->event_tree_results().size());
+  const auto& results = sequences();
+  ASSERT_EQ(1, results.size());
+  EXPECT_EQ("S", results.begin()->first);
+  EXPECT_DOUBLE_EQ(0.5, results.begin()->second);
+}
+
+TEST_P(RiskAnalysisTest, AnalyzeTestInitatingEvent) {
+  const char* tree_input = "./share/scram/input/eta/test_initiating_event.xml";
+  settings.probability_analysis(true);
+  ASSERT_NO_THROW(ProcessInputFile(tree_input));
+  ASSERT_NO_THROW(analysis->Analyze());
+  EXPECT_EQ(1, analysis->event_tree_results().size());
+  const auto& results = sequences();
+  ASSERT_EQ(1, results.size());
+  EXPECT_EQ("S", results.begin()->first);
+  EXPECT_DOUBLE_EQ(0.5, results.begin()->second);
+}
+
+TEST_P(RiskAnalysisTest, AnalyzeTestFunctionalEvent) {
+  const char* tree_input[] = {
+      "./share/scram/input/eta/test_functional_event.xml",
+      "./share/scram/input/eta/test_functional_event_link.xml"};
+  settings.probability_analysis(true);
+  for (auto input : tree_input) {
+    ASSERT_NO_THROW(ProcessInputFiles({input}));
+    ASSERT_NO_THROW(analysis->Analyze()) << input;
+    EXPECT_EQ(1, analysis->event_tree_results().size()) << input;
+    const auto& results = sequences();
+    ASSERT_EQ(1, results.size()) << input;
+    EXPECT_EQ("S", results.begin()->first) << input;
+    EXPECT_DOUBLE_EQ(0.5, results.begin()->second) << input;
+  }
+}
+
 // Test Reporting capabilities
 // Tests the output against the schema. However the contents of the
 // output are not verified or validated.
@@ -381,6 +476,11 @@ TEST_F(RiskAnalysisTest, ReportIOError) {
   EXPECT_THROW(Reporter().Report(*analysis, output), IOError);
 }
 
+TEST_F(RiskAnalysisTest, ReportEmpty) {
+  std::string tree_input = "./share/scram/input/empty_model.xml";
+  CheckReport(tree_input);
+}
+
 // Reporting of the default analysis for MCS only without probabilities.
 TEST_P(RiskAnalysisTest, ReportDefaultMCS) {
   CheckReport("./share/scram/input/fta/correct_tree_input.xml");
@@ -394,6 +494,18 @@ TEST_F(RiskAnalysisTest, ReportProbability) {
   CheckReport(tree_input);
 }
 
+TEST_F(RiskAnalysisTest, ReportProbabilityCurve) {
+  std::string tree_input = "./share/scram/input/core/single_exponential.xml";
+  settings.probability_analysis(true).time_step(24).mission_time(720);
+  CheckReport(tree_input);
+}
+
+TEST_F(RiskAnalysisTest, ReportSil) {
+  std::string tree_input = "./share/scram/input/core/single_exponential.xml";
+  settings.time_step(24).safety_integrity_levels(true).mission_time(720);
+  CheckReport(tree_input);
+}
+
 // Reporting of importance analysis.
 TEST_F(RiskAnalysisTest, ReportImportanceFactors) {
   std::string tree_input =
@@ -410,6 +522,13 @@ TEST_F(RiskAnalysisTest, ReportUncertaintyResults) {
   CheckReport(tree_input);
 }
 
+// Reporting event tree analysis with an initiating event.
+TEST_F(RiskAnalysisTest, ReportInitiatingEventAnalysis) {
+  const char* tree_input = "./share/scram/input/EventTrees/bcd.xml";
+  settings.probability_analysis(true);
+  CheckReport(tree_input);
+}
+
 // Reporting of CCF analysis.
 TEST_F(RiskAnalysisTest, ReportCCF) {
   std::string tree_input = "./share/scram/input/core/mgl_ccf.xml";
@@ -451,6 +570,11 @@ TEST_F(RiskAnalysisTest, ReportUnusedParameters) {
   CheckReport(tree_input);
 }
 
+TEST_F(RiskAnalysisTest, ReportUnusedEventTreeElements) {
+  std::string tree_input = "./share/scram/input/eta/unused_elements.xml";
+  CheckReport(tree_input);
+}
+
 // NAND and NOR as a child cases.
 TEST_P(RiskAnalysisTest, ChildNandNorGates) {
   std::string tree_input = "./share/scram/input/fta/children_nand_nor.xml";
